#!/usr/bin/env python

"""
What I've done here is create a bare-bones C-extension wrapper around the
lowlevel WPA socket, _WPACtrl, which I inherit from in Python to provide
more sensible interface in the form of the object WPACtrl.

Clients of this module are encouraged to ONLY use those classes whose
name does not begin with and underscore. They are truly unwieldy.

There are three tuples available in the global namespace: COMMANDS (the
recognized commands passed through the WPA socket), VARIABLES (the various
bit of data that wpa_supplicant associates with a "network"), and
EVENTS (a mostly non-helpful sequence of the "types" of events you can 
recieve).
"""

from _pywpa import _WPACtrl
from _pywpa import version
from _pywpa import error

import time

COMMANDS = (
	"PING",
	"MIB",
	"STATUS",
	"STATUS-VERBOSE",
	"PMKSA",
	"SET",
	"LOGON",
	"LOGOFF",
	"REASSOCIATE",
	"RREAUTH",
	"ATTACH",
	"DETACH",
	"LEVEL",
	"RECONFIGURE",
	"TERMINATE",
	"BSSID",
	"LIST_NETWORKS",
	"DISCONNECT",
	"SCAN",
	"SCAN_RESULTS",
	"SELECT_NETWORK",
	"ENABLE_NETWORK",
	"DISABLE_NETWORK",
	"ADD_NETWORK",
	"REMOVE_NETWORK",
	"SET_NETWORK",
	"GET_NETWORK",
	"SAVE_CONFIG"
)

VARIABLES = (
	"ssid",
	"disabled",
	"scan_ssid",
	"priority",
	"mode",
	"proto",
	"key_mgmt",
	"auth_alg",
	"pairwise",
	"group",
	"psk",
	"eapol_flags",
	"proactive_key_caching",
	"wep_key0",
	"wep_key1",
	"wep_key2",
	"wep_tx_keyidx"
)

EVENTS = (
	"CONNECTED",
	"DISCONNECTED",
	"TERMINATING",
	"PASSWORD_CHANGED",
	"EAP_NOTIFICATION",
	"EAP_STARTED",
	"EAP_METHOD",
	"EAP_SUCCESS",
	"EAP_FAILURE"
)

class _WPAEvent(object):
	def __init__(self, event):
		object.__init__(self)

		isevent = lambda e, t: e.startswith("%s-%s" % (
			"<2>CTRL-EVENT",
			t.replace("_", "-")
		))

		[setattr(self, ev.lower(), isevent(event, ev)) for ev in EVENTS]

		self.message = event[event.find(">") + 1 : ]

	def GetType(self):
		for event in EVENTS:
			if getattr(self, event.lower()):
				return event

	def __repr__(self):
		return self.message

class _WPANetwork(object):
	def __init__(self, initmap):
		object.__init__(self)

		[setattr(self, attr, "") for attr in VARIABLES]
		[setattr(self, attr, "") for attr in ("network_id", "bssid", "flags")]
		[setattr(self, attr, initmap[attr]) for attr in initmap]

		self.current = "[CURRENT]" in self.flags

	def __repr__(self):
		ret = "NETWORK (%s) / flags: %s\n\tcurrent -> %s\n\tbssid -> %s\n" % (
			self.network_id,
			self.flags,
			self.current,
			self.bssid
		)
		
		return ret + "".join(["\t%s -> %s\n" % (v, getattr(self, v)) for v in VARIABLES])
		
class WPACtrl(_WPACtrl):
	def __init__(self, path=""):
		run = "/var/run/wpa_supplicant"

		if not path:
			import os
		
			try:
				path = "%s/%s" % (run, os.listdir(run)[0])
			
			except OSError, error:
				path = "%s/eth1" % run

		elif not path.startswith("/"):
			path = "%s/%s" % (run, path)

		_WPACtrl.__init__(self, path)
	
	def GetEvent(self):
		return _WPAEvent(self.Recv())

	# TODO: When set to -1, this function needs to stop when communication with
	# the socket is dead. So, rar! I'm not sure how we'll do this, though...
	def GetEvents(self, num=-1):
		while num:
			event = self.GetEvent()

			if event.GetType():
				num -= 1
				
				yield event

			else:
				continue

	def GetPendingEvents(self):
		while self.Pending():
			event = self.GetEvent()

			if event.GetType():
				yield event

	def GetNetworks(self):
		wpanetworks = self.Request("LIST_NETWORKS")
		retnetworks = []
		maxtries    = 3

		# I'll want to investigate this later in C as well; if I can reproduce
		# it there it's not MY bug. :)
		while not wpanetworks.lstrip().rstrip().replace("\n", "") and maxtries:
			wpanetworks = self.Request("LIST_NETWORKS")
			maxtries   -= 1

		if not maxtries:
			raise error, "Couldn't get networks; tried 3 times."

		wpanetworks = wpanetworks.split("\n")[1 : -1]

		for net in range(len(wpanetworks)):
			network = wpanetworks[net].split()
			initmap = {}

			[network.append("") for i in range(4 - len(network))]

			initmap["network_id"] = network[0]
			initmap["bssid"]      = network[2]
			initmap["flags"]      = "".join(network[3 : ])

			for v in VARIABLES:
				response   = self.Request("GET_NETWORK %s %s" % (network[0], v))
				initmap[v] = "FAIL" in response and "UNSET" or response

			retnetworks.append(_WPANetwork(initmap))

		return retnetworks

	def SetNetwork(self, net, param, val):
		return not "FAIL" in self.Request("SET_NETWORK %s %s %s" % (net, param, val))

	# This is a basic helper for adding a new network relatively quickly.
	def AddNetwork(self, ssid, pri=None, psk=None, wep=None):
		net = self.Request("ADD_NETWORK").replace("\n", "")

		if not "FAIL" in net:
			if self.SetNetwork(net, "ssid", '"%s"' % ssid):
				#self.Request("ENABLE_NETWORK %s" % net)
			
				if psk:
					self.SetNetwork(net, "psk", '"%s"' % psk)
					self.SetNetwork(net, "key_mgmt", "WPA-PSK")
				else:
					self.SetNetwork(net, "key_mgmt", "NONE")
				
				wep and self.SetNetwork(net, "wep_key0", wep)
				pri and self.SetNetwork(net, "priority", pri)

			else:
				self.Request("REMOVE_NETWORK %s" % net)
				
				raise error, "Coudn't set network SSID."

		else:
			raise error, "Couldn't create new network."

	def ActivateNetwork(self, network_id):
		if self.Request("SELECT_NETWORK %s" % network_id).startswith("FAIL"):
			raise error, "Couldn't activate network %d" % network_id
	
	def Reassociate(self):
		if self.Request("REASSOCIATE").startswith("FAIL"):
			raise error, "Couldn't reassociate."

	def DeleteNetwork(self, network_id):
		if self.Request("REMOVE_NETWORK %s" % network_id).startswith("FAIL"):
			raise error, "Couldn't delete network %d" % network_id

	def SaveNetworks(self):
		if self.Request("SAVE_CONFIG").startswith("FAIL"):
			raise error, "Couldn't save networks."

	def EnableAllNetworks(self):
		nets = self.Request("LIST_NETWORKS")

		for net in range(1, len(nets) + 1):
			self.Request("ENABLE_NETWORK %d" % net)

	def IsAlive(self):
		try:
			self.Request("PING")

			return True

		except error:
			return False
